import functools
import keyword
import warnings
from collections import defaultdict

import numpy
import xarray
from tiled.client.composite import Composite
from tiled.client.container import DEFAULT_STRUCTURE_CLIENT_DISPATCH, Container
from tiled.utils import DictView, OneShotCachedMap, Sentinel, node_repr

from ._common import IPYTHON_METHODS

DATAVALUES = Sentinel("DATAVALUES")
TIMESTAMPS = Sentinel("TIMESTAMPS")


class BlueskyEventStream(Container):
    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting from BlueskyEventStream, return the class itself
        if cls is not BlueskyEventStream:
            return super().__new__(cls)

        # Set the version based on the specs
        _cls = BlueskyEventStreamV3 if cls._is_sql(item) else BlueskyEventStreamV2Mongo
        return _cls(context, item=item, structure_clients=structure_clients, **kwargs)

    @staticmethod
    def _is_sql(item):
        for spec in item["attributes"]["specs"]:
            if spec["name"] == "BlueskyEventStream":
                if spec["version"].startswith("3."):
                    return True
                return False


class BlueskyEventStreamV2Mongo(BlueskyEventStream):
    """
    This encapsulates the data and metadata for one 'stream' in a Bluesky 'run'.

    This adds for bluesky-specific conveniences to the standard client Container.
    """

    def __repr__(self):
        stream_name = self.metadata.get("stream_name") or self.item["id"]
        return f"<BlueskyEventStream {set(self)!r} stream_name={stream_name!r}>"

    @property
    def descriptors(self):
        return self.metadata["descriptors"]

    @property
    def _descriptors(self):
        # For backward-compatibility.
        # We do not normally worry about backward-compatibility of _ methods, but
        # for a time databroker.v2 *only* have _descriptors and not descriptors,
        # and I know there is useer code that relies on that.
        warnings.warn("Use `.descriptors` instead of `._descriptors`.", stacklevel=2)
        return self.descriptors

    def __getattr__(self, key):
        """
        Let run.X be a synonym for run['X'] unless run.X already exists.

        This behavior is the same as with pandas.DataFrame.
        """
        # The wisdom of this kind of "magic" is arguable, but we
        # need to support it for backward-compatibility reasons.
        if key in IPYTHON_METHODS:
            raise AttributeError(key)
        if key in self:
            return self[key]
        raise AttributeError(key)

    def __dir__(self):
        # Build a list of entries that are valid attribute names
        # and add them to __dir__ so that they tab-complete.
        tab_completable_entries = [
            entry for entry in self if (entry.isidentifier() and (not keyword.iskeyword(entry)))
        ]
        return super().__dir__() + tab_completable_entries

    def read(self, *args, **kwargs):
        """
        Shortcut for reading the 'data' (as opposed to timestamps or config).

        That is:

        >>> stream.read(...)

        is equivalent to

        >>> stream["data"].read(...)
        """
        return self["data"].read(*args, **kwargs)

    def to_dask(self):
        warnings.warn(
            """Do not use this method.
Instead, set dask or when first creating the client, as in

    >>> catalog = from_uri("...", "dask")

and then read() will return dask objects.""",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.new_variation(structure_clients=DEFAULT_STRUCTURE_CLIENT_DISPATCH["dask"]).read()


class BlueskyEventStreamV2SQL(OneShotCachedMap):
    def __init__(self, internal_dict, metadata=None):
        super().__init__(internal_dict)
        self.metadata = metadata or {}

    def __repr__(self):
        stream_name = self.metadata.get("stream_name")
        return f"<BlueskyEventStream {set(self)!r} stream_name={stream_name!r}>"

    def __getitem__(self, key):
        if "/" in key:
            key, rest = key.split("/", 1)
            return self[key][rest]

        return super().__getitem__(key)

    @classmethod
    def from_stream_client(cls, stream_client, metadata=None):
        stream_parts = set(stream_client.parts)
        data_keys = [k for k in stream_parts if k != "internal"]
        ts_keys = ["time"]
        if "internal" in stream_parts:
            internal_cols = stream_client.parts["internal"].columns
            data_keys += [col for col in internal_cols if col != "seq_num" and not col.startswith("ts_")]
            ts_keys += [col for col in internal_cols if col.startswith("ts_")]

        # Construct clients for the configuration data
        cf_vals, cf_time = defaultdict(dict), defaultdict(dict)
        if config := stream_client.metadata.get("configuration", {}):
            updates = stream_client.metadata.get("_config_updates", [])
            for obj_name, obj in config.items():
                for key in obj["data"].keys():
                    _vs, _ts = [obj["data"][key]], [obj["timestamps"][key]]

                    # Add values and timestamps from config_updates
                    for upd in updates:
                        if upd_config := upd.get("configuration", {}):
                            _vs.append(upd_config.get("data", {}).get(key))
                            _ts.append(upd_config.get("timestamps", {}).get(key))

                    cf_vals[obj_name][key] = VirtualArrayClient(_vs)
                    cf_time[obj_name][key] = VirtualArrayClient(_ts)

        internal_dict = {
            "data": lambda: CompositeSubsetClient(stream_client, data_keys),
            "timestamps": lambda: CompositeSubsetClient(stream_client, ts_keys),
            "config": lambda: VirtualContainer({k: ConfigDatasetClient(v) for k, v in cf_vals.items()}),
            "config_timestamps": lambda: VirtualContainer({k: ConfigDatasetClient(v) for k, v in cf_time.items()}),
        }

        # Construct the metadata
        metadata = {
            "descriptors": [],
            "stream_name": stream_client.item["id"],
            **stream_client.metadata,
            **(metadata or {}),
        }

        return cls(internal_dict, metadata=metadata)

    @functools.cached_property
    def descriptors(self):
        # Go back to the BlueskyRun node and requests the documents
        bs_run_node = self["data"].parent.parent  # the path is: bs_run_node/streams/current_stream
        stream_name = self.metadata.get("stream_name") or self["data"].item["id"]
        return [
            doc for name, doc in bs_run_node.documents() if name == "descriptor" and doc["name"] == stream_name
        ]

    @property
    def _descriptors(self):
        # For backward-compatibility.
        # We do not normally worry about backward-compatibility of _ methods, but
        # for a time databroker.v2 *only* have _descriptors and not descriptors,
        # and I know there is useer code that relies on that.
        warnings.warn("Use `.descriptors` instead of `._descriptors`.", stacklevel=2)
        return self.descriptors

    def __getattr__(self, key):
        """
        Let run.X be a synonym for run['X'] unless run.X already exists.

        This behavior is the same as with pandas.DataFrame.
        """
        # The wisdom of this kind of "magic" is arguable, but we
        # need to support it for backward-compatibility reasons.
        if key in IPYTHON_METHODS:
            raise AttributeError(key)
        if key in self:
            return self[key]
        raise AttributeError(key)

    def read(self, *args, **kwargs):
        """Read the data from the stream.

        This is a shortcut for reading the 'data' (as opposed to timestamps or config).
        """
        return self["data"].read(*args, **kwargs)


class ConfigDatasetClient(DictView):
    def __repr__(self):
        tiled_repr = node_repr(self, self._internal_dict.keys())
        return tiled_repr.replace(type(self).__name__, "DatasetClient")

    def read(self):
        # Delay this import for fast startup. In some cases only metadata
        # is handled, and we can avoid the xarray import altogether.

        d = {k: {"dims": "time", "data": v.read()} for k, v in self._internal_dict.items()}
        return xarray.Dataset.from_dict(d)


class CompositeSubsetClient(Composite):
    """A composite client with only a subset of its keys exposed."""

    def __init__(self, client, keys=None):
        super().__init__(context=client.context, item=client.item, structure_clients=client.structure_clients)
        self._keys = keys or list(client.keys())

    def __repr__(self):
        return node_repr(self, self._keys).replace(type(self).__name__, "DatasetClient")

    def _keys_slice(self, start, stop, direction, _ignore_inlined_contents=False):
        yield from self._keys[start : stop : -1 if direction < 0 else 1]  # noqa: #203

    def _items_slice(self, start, stop, direction, _ignore_inlined_contents=False):
        for key in self._keys[start : stop : -1 if direction < 0 else 1]:  # noqa: #203
            yield key, self[key]

    def __iter__(self):
        yield from self._keys

    def __getitem__(self, key):
        if key in self._keys:
            return super().__getitem__(key)
        raise KeyError(key)

    def __len__(self):
        return len(self._keys)

    def __contains__(self, key):
        return key in self._keys

    def read(self, variables=None, dim0=None):
        variables = set(self._keys).intersection(variables or self._keys)

        return super().read(variables, dim0=dim0)


class VirtualContainer(DictView):
    def __repr__(self):
        tiled_repr = node_repr(self, self._internal_dict.keys())
        return tiled_repr.replace(type(self).__name__, "ContainerClient")

    def __getitem__(self, key):
        if "/" in key:
            key, rest = key.split("/", 1)
            return self[key][rest]

        return super().__getitem__(key)


class VirtualArrayClient:
    def __init__(self, data, dims=None):
        # Delay this import for fast startup. In some cases only metadata
        # is handled, and we can avoid the numpy import altogether.

        # Ensure data is an array-like object
        if not hasattr(data, "__iter__") or isinstance(data, str):
            data = [data]
        if not hasattr(data, "__array__"):
            data = numpy.asanyarray(data)

        self._data = data
        self._dims = dims

    def __getitem__(self, slice):
        return self.read(slice)

    def __repr__(self):
        attrs = {"shape": self.shape, "dtype": self.dtype}
        if dims := self.dims:
            attrs["dims"] = dims
        return "<ArrayClient" + "".join(f" {k}={v}" for k, v in attrs.items()) + ">"

    def read(self, slice=None):
        return self._data if slice is None else self._data[slice]

    @property
    def size(self):
        return self._data.size

    @property
    def shape(self):
        return self._data.shape

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def dims(self):
        return self._dims


class BlueskyEventStreamV3(BlueskyEventStream, Composite):
    def __repr__(self):
        stream_name = self.metadata.get("stream_name") or self.item["id"]
        return f"<BlueskyEventStream {self._var_keys!r} stream_name={stream_name!r}>"

    @property
    def _var_keys(self):
        return {k for k in self if not k.startswith("ts_") and k != "seq_num"}

    @property
    def _ts_keys(self):
        return {k for k in self if k.startswith("ts_")}

    def read(self, variables=(DATAVALUES,), dim0=None):
        if DATAVALUES in variables:
            variables = self._var_keys.union(variables) - {DATAVALUES}
        if TIMESTAMPS in variables:
            variables = self._ts_keys.union(variables) - {TIMESTAMPS}

        return super().read(variables=variables, dim0=dim0)

    @functools.cached_property
    def descriptors(self):
        # Go back to the BlueskyRun node and requests the documents
        stream_name = self.metadata.get("stream_name") or self.item["id"]
        bs_run_node = self.parent.parent  # the path is: bs_run_node/streams/current_stream
        return [
            doc for name, doc in bs_run_node.documents() if name == "descriptor" and doc["name"] == stream_name
        ]



import copy

from dask.base import normalize_token


class NotMutable(Exception):
    pass


class Document(dict):
    """
    Document is an immutable dict subclass.

    It is immutable to help consumer code avoid accidentally corrupting data
    that another part of the consumer code was expected to use unchanged.

    Subclasses of Document must define __dask_tokenize__. The tokenization
    schemes typically uniquely identify the document based on only a subset of
    its contents, and mutating the contents can thereby create situations where
    two unequal objects have colliding tokens. Immutability helps guard against
    this too.

    Note that Documents are not *recursively* immutable. Just as it is possible
    create a tuple (immutable) of lists (mutable) and mutate the lists, it is
    possible to mutate the internal contents of a Document, but this should not
    be done. It is safer to use the to_dict() method to create a mutable deep
    copy.

    This is implemented as a dict subclass in order to satisfy certain
    consumers that expect an object that satisfies isinstance(obj, dict).
    This implementation detail may change in the future.
    """

    __slots__ = ("__not_a_real_dict",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # This lets pickle recognize that this is not a literal dict and that
        # it should respect its custom __setstate__.
        self.__not_a_real_dict = True

    def __repr__(self):
        # same as dict, but wrapped in the class name so the eval round-trips
        return f"{self.__class__.__name__}({dict(self)})"

    def _repr_pretty_(self, p, cycle):
        """
        A multi-line but eval-able text repr with readable indentation

        This hooks into IPython/Jupyter's display mechanism
        This is *not* invoked by print() or repr(), but it is invoked by
        IPython.display.display() which is called in this common scenario::

            In [1]: doc = Document(...)
            In [2]: doc
            <pretty representation will show here>
        """
        # Note: IPython's pretty-prettying mechanism is custom and complex.
        # The `text` method used below is a direct and blunt way to engage it
        # and seems widely used in the IPython code base. There are other
        # specific mechanisms for displaying collections like dicts, but they
        # can *truncate* which I think we want to avoid and they would require
        # more investment to understand how to use.
        from pprint import pformat

        return p.text(f"{self.__class__.__name__}({pformat(dict(self))})")

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, state):
        dict.update(self, state)
        self.__not_a_real_dict = True

    def __readonly(self, *args, **kwargs):
        raise NotMutable(
            "Documents are not mutable. Call the method to_dict() to make a "
            "fully independent and mutable deep copy."
        )

    def __setitem__(self, key, value):
        try:
            self.__not_a_real_dict  # noqa: B018
        except AttributeError:
            # This path is necessary to support un-pickling.
            return dict.__setitem__(self, key, value)
        else:
            self.__readonly()

    __delitem__ = __readonly
    pop = __readonly
    popitem = __readonly
    clear = __readonly
    setdefault = __readonly
    update = __readonly

    def to_dict(self):
        """
        Create a mutable deep copy.
        """
        # Convert to dict and then make a deep copy to ensure that if the user
        # mutates any internally nested dicts there is no spooky action at a
        # distance.
        return copy.deepcopy(dict(self))

    def __deepcopy__(self, memo):
        # Without this, copy.deepcopy(Document(...)) fails because deepcopy
        # creates a new, empty Document instance and then tries to add items to
        # it.
        return self.__class__({k: copy.deepcopy(v, memo) for k, v in self.items()})

    def __dask_tokenize__(self):
        raise NotImplementedError


# We must use dask's registration mechanism to tell it to treat Document
# specially. Dask's tokenization dispatch mechanism discovers that Docuemnt is
# a dict subclass and treats it as a dict, ignoring its __dask_tokenize__
# method. To force it to respect our cutsom tokenization, we must explicitly
# register it.


@normalize_token.register(Document)
def tokenize_document(instance):
    return instance.__dask_tokenize__()


class Start(Document):
    def __dask_tokenize__(self):
        return ("start", self["uid"])


class Stop(Document):
    def __dask_tokenize__(self):
        return ("stop", self["uid"])


class Resource(Document):
    def __dask_tokenize__(self):
        return ("resource", self["uid"])


class StreamResource(Document):
    def __dask_tokenize__(self):
        return ("stream_resource", self["uid"])


class Descriptor(Document):
    def __dask_tokenize__(self):
        return ("descriptor", self["uid"])


class Event(Document):
    def __dask_tokenize__(self):
        return ("event", self["uid"])


class EventPage(Document):
    def __dask_tokenize__(self):
        return ("event_page", self["uid"])


class Datum(Document):
    def __dask_tokenize__(self):
        return ("datum", self["datum_id"])


class StreamDatum(Document):
    def __dask_tokenize__(self):
        return ("stream_datum", self["uid"])


class DatumPage(Document):
    def __dask_tokenize__(self):
        return ("datum_page", self["uid"])



import copy
import functools
import io
import json
import keyword
import warnings
from datetime import datetime

from tiled.client.container import Container
from tiled.client.utils import handle_error

from ._common import IPYTHON_METHODS
from .bluesky_event_stream import BlueskyEventStreamV2SQL
from .document import DatumPage, Descriptor, Event, EventPage, Resource, Start, Stop, StreamDatum, StreamResource

_document_types = {
    "start": Start,
    "stop": Stop,
    "event": Event,
    "descriptor": Descriptor,
    "event_page": EventPage,
    "datum_page": DatumPage,
    "resource": Resource,
    "stream_resource": StreamDatum,
    "stream_datum": StreamResource,
}

RESERVED_V3_KEYS = {"configs", "streams", "views", "aux"}


class BlueskyRun(Container):
    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting from BlueskyRun, return the class itself
        if cls is not BlueskyRun:
            return super().__new__(cls)

        # Set the version based on the specs
        _cls = BlueskyRunV3 if cls._is_sql(item) else BlueskyRunV2Mongo
        return _cls(context, item=item, structure_clients=structure_clients, **kwargs)

    @staticmethod
    def _is_sql(item):
        for spec in item["attributes"]["specs"]:
            if spec["name"] == "BlueskyRun":
                if spec["version"].startswith("3."):
                    return True
                return False

    def __repr__(self):
        metadata = self.metadata
        datetime_ = datetime.fromtimestamp(metadata["start"]["time"])
        return (
            f"<BlueskyRun v{self._version} "
            f"{set(self)!r} "  # show the keys
            f"scan_id={metadata['start'].get('scan_id', 'UNSET')!s} "  # (scan_id is optional in the schema)
            f"uid={metadata['start']['uid'][:8]!r} "  # truncated uid
            f"{datetime_.isoformat(sep=' ', timespec='minutes')}"
            ">"
        )

    @property
    def start(self):
        """
        The Run Start document. A convenience alias:

        >>> run.start is run.metadata["start"]
        True
        """
        return self.metadata["start"]

    @property
    def stop(self):
        """
        The Run Stop document. A convenience alias:

        >>> run.stop is run.metadata["stop"]
        True
        """
        return self.metadata["stop"]

    @functools.cached_property
    def descriptors(self):
        return [doc for name, doc in self.documents() if name == "descriptor"]

    def __getattr__(self, key):
        """
        Let run.X be a synonym for run['X'] unless run.X already exists.

        This behavior is the same as with pandas.DataFrame.
        """
        # The wisdom of this kind of "magic" is arguable, but we
        # need to support it for backward-compatibility reasons.
        if key in IPYTHON_METHODS:
            raise AttributeError(key)
        if key in self:
            return self[key]
        raise AttributeError(key)

    def __dir__(self):
        # Build a list of entries that are valid attribute names
        # and add them to __dir__ so that they tab-complete.
        tab_completable_entries = [
            entry for entry in self if (entry.isidentifier() and (not keyword.iskeyword(entry)))
        ]
        return super().__dir__() + tab_completable_entries

    def describe(self):
        "For back-compat with intake-based BlueskyRun"
        warnings.warn(
            "This will be removed. Use .metadata directly instead of describe()['metadata'].",
            DeprecationWarning,
            stacklevel=2,
        )
        return {"metadata": self.metadata}

    def __call__(self):
        warnings.warn(
            "Do not call a BlueskyRun. For now this returns self, for "
            "backward-compatibility. but it will be removed in a future "
            "release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    def read(self):
        raise NotImplementedError(
            "Reading any entire run is not supported. Access a stream in this run and read that."
        )

    to_dask = read


class BlueskyRunV2(BlueskyRun):
    """A MongoDB-native layout of BlueskyRuns

    This layout has been in use prior to the introduction of SQL backend in May 2025.
    """

    _version = "2.0"

    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting, return the class itself
        if cls is not BlueskyRunV2:
            return super().__new__(cls, context, item=item, structure_clients=structure_clients, **kwargs)

        _cls = BlueskyRunV2SQL if cls._is_sql(item) else BlueskyRunV2Mongo
        return _cls(context, item=item, structure_clients=structure_clients, **kwargs)

    @property
    def v1(self):
        "Accessor to legacy interface."
        from databroker.v1 import Broker, Header

        db = Broker(self)
        header = Header(self, db)
        return header

    @property
    def v2(self):
        return self

    @property
    def v3(self):
        if not self._is_sql(self.item):
            raise NotImplementedError("v3 is not available for MongoDB-based BlueskyRun")

        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV3)
        return BlueskyRunV3(self.context, item=self.item, structure_clients=structure_clients)


class BlueskyRunV2Mongo(BlueskyRunV2):
    def documents(self, fill=False):
        if fill == "yes":
            fill = True
        elif fill == "no":
            fill = False
        elif fill == "delayed":
            raise NotImplementedError("fill='delayed' is not supported")
        else:
            fill = bool(fill)
        link = self.item["links"]["self"].replace("/metadata", "/documents", 1)
        with self.context.http_client.stream(
            "GET",
            link,
            params={"fill": fill},
            headers={"Accept": "application/json-seq"},
        ) as response:
            if response.is_error:
                response.read()
                handle_error(response)
            tail = ""
            for chunk in response.iter_bytes():
                for line in chunk.decode().splitlines(keepends=True):
                    if line[-1] == "\n":
                        item = json.loads(tail + line)
                        yield (item["name"], _document_types[item["name"]](item["doc"]))
                        tail = ""
                    else:
                        tail += line
            if tail:
                item = json.loads(tail)
                yield (item["name"], _document_types[item["name"]](item["doc"]))


class _BlueskyRunSQL(BlueskyRun):
    """A base class for a BlueskyRun that is backed by a SQL database.

    This class implements the SQL-specific method for accessing the stream of
    Bluesky documents. It is not intended to be used directly, but rather as a
    base class for other classes (v2 and v3) that implement additional methods.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @functools.cached_property
    def _stream_names(self):
        return sorted(self.get("streams", ()))

    def documents(self, fill=False):
        with io.BytesIO() as buffer:
            self.export(buffer, format="application/json-seq")
            buffer.seek(0)
            for line in buffer:
                parsed = json.loads(line.decode().strip())
                yield parsed["name"], _document_types[parsed["name"]](parsed["doc"])


class BlueskyRunV2SQL(BlueskyRunV2, _BlueskyRunSQL):
    def _keys_slice(self, start, stop, direction, **kwargs):
        keys = reversed(self._stream_names) if direction < 0 else self._stream_names
        return (yield from keys[start:stop])

    def _items_slice(self, start, stop, direction, **kwargs):
        _streams_node = super().get("streams", {})
        for key in reversed(self._stream_names) if direction < 0 else self._stream_names:
            yield key, _streams_node.get(key)
        return

    def __getitem__(self, key):
        # For v3, we need to handle the streams and configs keys
        if key in RESERVED_V3_KEYS:
            return super().__getitem__(key)

        if key in self._stream_names:
            stream_container = super().get("streams", {}).get(key)
            return BlueskyEventStreamV2SQL.from_stream_client(stream_container)

        if "/" in key:
            key, rest = key.split("/", 1)
            return self[key][rest]

        return super().__getitem__(key)

    def __iter__(self):
        yield from self._stream_names


class BlueskyRunV3(_BlueskyRunSQL):
    """A BlueskyRun that is backed by a SQL database."""

    _version = "3.0"

    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting, return the class itself
        if cls is not BlueskyRunV3 or cls._is_sql(item):
            return super().__new__(cls, context, item=item, structure_clients=structure_clients, **kwargs)
        else:
            return BlueskyRunV2Mongo(context, item=item, structure_clients=structure_clients, **kwargs)

    def __getattr__(self, key):
        if key in self._stream_names:
            # A shortcut to the stream data
            return self["streams"][key]

        return super().__getattr__(key)

    def __repr__(self):
        metadata = self.metadata
        datetime_ = datetime.fromtimestamp(metadata["start"]["time"])
        return (
            f"<BlueskyRun v{self._version} "
            f"streams: {set(self._stream_names) or 'NONE'} "
            f"scan_id={metadata['start'].get('scan_id', 'UNSET')!s} "  # (scan_id is optional in the schema)
            f"uid={metadata['start']['uid'][:8]!r} "  # truncated uid
            f"{datetime_.isoformat(sep=' ', timespec='minutes')}"
            ">"
        )

    @property
    def v1(self):
        "Access to legacy interface"
        return self.v2.v1

    @property
    def v2(self):
        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV2)
        return BlueskyRunV2(self.context, item=self.item, structure_clients=structure_clients)

    @property
    def v3(self):
        return self



import copy
import json
from collections import defaultdict


async def json_seq_exporter(adapter, metadata, filter_for_access):
    """
    Export BlueskyRun is newline-delimited sequence of JSON.

    Format is like:

    {"name": "start", "doc": {...}}
    {"name": "descriptor", "doc": {...}}
    ...
    """
    for spec in adapter.specs:
        if spec.name == "BlueskyRun" and spec.version.startswith("3."):
            break
    else:
        raise ValueError("This exporter only works with BlueskyRun v3.x")

    adapter = await filter_for_access(adapter, path_parts=[])
    yield json.dumps({"name": "start", "doc": metadata.get("start", {})})
    result = []

    # Generate descriptors
    stream_names = await (await adapter.lookup_adapter(["streams"])).keys_range(offset=0, limit=None)
    for desc_name in stream_names:
        desc_node = await adapter.lookup_adapter(["streams", desc_name])
        desc_meta = desc_node.metadata()
        part_names = set(await desc_node.keys_range(offset=0, limit=None))  # Composite parts

        # First (or the only) descriptor
        desc_doc = {k: v for k, v in desc_meta.items() if k not in {"_config_updates"}}
        desc_doc["run_start"] = metadata.get("start", {}).get("uid")
        desc_doc["name"] = desc_name
        desc_doc["object_keys"] = defaultdict(list)
        for key, val in desc_doc["data_keys"].items():
            if obj_name := val.get("object_name"):
                desc_doc["object_keys"][obj_name].append(key)

        result.append({"name": "descriptor", "doc": desc_doc})

        # Process subsequent descriptors, if any
        desc_time_uids = [{"uid": desc_doc["uid"], "time": desc_doc["time"]}]
        for upd in desc_meta.get("_config_updates", []):
            desc_doc = copy.deepcopy(desc_doc)
            desc_doc["uid"] = upd["uid"]
            desc_doc["time"] = upd["time"]
            desc_time_uids.extend([{"uid": desc_doc["uid"], "time": desc_doc["time"]}])
            for obj_name, obj in upd.get("configuration", {}).items():
                # This assumes that that the full configuration was present in the first descriptor
                for key in obj["data"].keys():
                    desc_doc["configuration"][obj_name]["data"][key] = obj["data"][key]
                    desc_doc["configuration"][obj_name]["timestamps"][key] = obj["timestamps"][key]

            result.append({"name": "descriptor", "doc": desc_doc})

        # Generate events
        if "internal" in part_names:
            internal_node = await desc_node.lookup_adapter(["internal"])
            df = await internal_node.read()
            keys = [k for k in df.columns if k not in {"seq_num", "time"} and not k.startswith("ts_")]
            for row in df.to_dict(orient="records"):
                desc_uid = desc_time_uids[0]["uid"]  # same as desc_node.metadata()["uid"] if no updates
                for _desc_uid_time in desc_time_uids[1:]:
                    if _desc_uid_time["time"] <= row["time"]:
                        desc_uid = _desc_uid_time["uid"]
                event_doc = {"seq_num": row["seq_num"], "time": row["time"]}
                event_doc["uid"] = f"event-{desc_uid}-{row['seq_num']}"  # can be anything (unique)
                event_doc["descriptor"] = desc_uid
                event_doc["data"] = {k: row[k].tolist() if hasattr(row[k], "__array__") else row[k] for k in keys}
                event_doc["timestamps"] = {k: row[f"ts_{k}"] for k in keys}
                result.append({"name": "event", "doc": event_doc})

        # Generate Stream Resources and Datums
        desc_uid = desc_node.metadata()["uid"]
        for data_key in part_names.difference(("internal",)):
            # Loop over data_keys for external data only
            sres_uid = f"sr-{desc_uid}-{data_key}"  # can be anything (unique)
            ds = (await desc_node.lookup_adapter([data_key])).data_sources[0]
            uri = ds.assets[0].data_uri
            for ast in ds.assets:
                if ast.parameter in {"data_uris", "data_uri"}:
                    uri = ast.data_uri
                    break
            sres_doc = {
                "data_key": data_key,
                "uid": sres_uid,
                "run_start": metadata.get("start", {}).get("uid"),
                "mimetype": ds.mimetype,
                "parameters": ds.parameters,
                "uri": uri,
            }
            result.append({"name": "stream_resource", "doc": sres_doc})

            # Generate a single stream_datum document for the entire stream
            sdat_uid = f"sd-{desc_uid}-{data_key}-0"  # can be anything (unique)
            total_shape = ds.structure.shape
            datum_shape = desc_node.metadata()["data_keys"][data_key]["shape"]

            max_indx = (
                total_shape[0] // datum_shape[0] - 1
                if len(total_shape) == len(datum_shape)
                else total_shape[0] - 1
            )
            sdat_doc = {
                "uid": sdat_uid,
                "stream_resource": sres_uid,
                "descriptor": desc_uid,
                "indices": {"start": 0, "stop": max_indx},
                "seq_nums": {"start": 1, "stop": max_indx + 1},
            }
            result.append({"name": "stream_datum", "doc": sdat_doc})

    # Make sure that the order of documents is (approximately) correct
    result = sorted(
        result,
        key=lambda x: (
            x["doc"].get("time", float("inf")),
            {"stream_resource": 0, "stream_datum": 1}.get(x["name"]),
        ),
    )

    # Combine events into event_pages
    #     if modules_available("databroker"):
    #         from databroker.mongo_normalized import batch_documents
    #
    #         result = [
    #             {"name": x[0], "doc": x[1]}
    #             for x in batch_documents([(y["name"], y["doc"]) for y in result], size=1000)
    #         ]

    for doc in result:
        yield "\n" + json.dumps(doc)

    yield "\n" + json.dumps({"name": "stop", "doc": metadata.get("stop", {})})



# There are methods that IPython will try to call.
# We special-case them because we want to avoid the getattr
# resulting in an unnecessary network hit just to raise
# AttributeError.

IPYTHON_METHODS = {"_ipython_canary_method_should_not_exist_", "_repr_mimebundle_"}



"""
This module defines objects designed to make queries on
CatalogOfBlueskyRuns convenient.

Older clients used these query object to issue custom query types.
This requires servers to register custom implementations of those
query types.

Newer clients use these object as pure client-side conveniences. In
`CatalogOfBlueskyRuns.search` method, they are decomposed into standard Tiled
queries, requiring no custom counterpart on the server.

The registration and serialization aspects are (temporarily) retained in order
to support older clients querying against MongoDB-backed servers.
"""

import enum
import warnings
from dataclasses import asdict, dataclass
from typing import Optional

# Not all of these are used, but import them all
# for user convenience so everything can be imported from bluesky_tiled_plugins.queries
from tiled.queries import (  # noqa: F401
    Comparison,
    Contains,
    Eq,
    FullText,
    In,
    Key,
    NotEq,
    NotIn,
    Operator,
    QueryValueError,
    Regex,
)
from tiled.query_registration import register


class Duplicates(str, enum.Enum):
    latest = "latest"
    all = "all"
    error = "error"


@register(name="scan_id")
@dataclass
class _ScanID:
    """
    Find matches to scan_id(s).
    """

    scan_ids: list[int]
    duplicates: Duplicates

    def __init__(self, *, scan_ids, duplicates):
        self.scan_ids = scan_ids
        self.duplicates = Duplicates(duplicates)

    def encode(self):
        return {
            "scan_ids": ",".join(str(scan_id) for scan_id in self.scan_ids),
            "duplicates": self.duplicates.value,
        }

    @classmethod
    def decode(cls, *, scan_ids, duplicates):
        return cls(
            scan_ids=[int(scan_id) for scan_id in scan_ids.split(",")],
            duplicates=Duplicates(duplicates),
        )


def ScanID(*scan_ids, duplicates="latest"):
    # Wrap _ScanID to provide a nice usage for *one or more scan_ids*:
    # >>> ScanID(5)
    # >>> ScanID(5, 6, 7)
    # Placing a varargs parameter (*scan_ids) in the dataclass constructor
    # would cause trouble on the server side and generally feels "wrong"
    # so we have this wrapper function instead.
    return _ScanID(scan_ids=scan_ids, duplicates=duplicates)


@register(name="scan_id_range")
@dataclass
class ScanIDRange:
    """
    Find scans in the range.
    """

    start_id: int
    end_id: int
    duplicates: Duplicates

    def __init__(self, start_id, end_id, duplicates="latest"):
        self.start_id = start_id
        self.end_id = end_id
        self.duplicates = Duplicates(duplicates)

    def encode(self):
        return {
            "start_id": self.start_id,
            "end_id": self.end_id,
            "duplicates": self.duplicates.value,
        }

    @classmethod
    def decode(cls, *, start_id, end_id, duplicates="latest"):
        return cls(
            start_id=int(start_id),
            end_id=int(end_id),
            duplicates=Duplicates(duplicates),
        )


@register(name="partial_uid")
@dataclass
class _PartialUID:
    """
    Find matches to (partial) uid(s).
    """

    partial_uids: list[str]

    def encode(self):
        return {"partial_uids": ",".join(str(uid) for uid in self.partial_uids)}

    @classmethod
    def decode(cls, *, partial_uids):
        return cls(partial_uids=partial_uids.split(","))


def PartialUID(*partial_uids):
    # See comment above with ScanID and _ScanID. Same thinking here.
    return _PartialUID(partial_uids)


def RawMongo(start):
    """
    DEPRECATED

    Raw MongoDB queries are no longer supported.  If it is possible to express
    the import as a supported query, we transform it and warn. If not, we raise
    an error.
    """

    if len(start) == 1:
        ((key, value),) = start.items()
        if not isinstance(value, dict):
            # We can transform this into a simple query.
            warnings.warn(
                """RawMongo will not be supported
in a future release of databroker, and its functionality has been limited.
Instead, use:

    Key("{key}") == {value!r}
""",
                stacklevel=2,
            )
            return Key(key) == value
    raise ValueError(
        """Arbitrary MongoDB queries no longer supported.

If this is critical to you, please open an issue at

    https://github.com/bluesky/databroker

describing your use case and we will see what we can work out."""
    )


# human friendly timestamp formats we'll parse
_TS_FORMATS = [
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M",  # these 2 are not as originally doc'd,
    "%Y-%m-%d %H",  # but match previous pandas behavior
    "%Y-%m-%d",
    "%Y-%m",
    "%Y",
]


def _normalize_human_friendly_time(val, tz):
    """Given one of :
    - string (in one of the formats below)
    - datetime (eg. datetime.now()), with or without tzinfo)
    - timestamp (eg. time.time())
    return a timestamp (seconds since jan 1 1970 UTC).

    Non string/datetime values are returned unaltered.
    Leading/trailing whitespace is stripped.
    Supported formats:
    {}
    """
    # {} is placeholder for formats; filled in after def...

    from datetime import datetime

    import pytz

    zone = pytz.timezone(tz)  # tz as datetime.tzinfo object
    epoch = pytz.UTC.localize(datetime(1970, 1, 1))
    check = True

    if isinstance(val, str):
        # unix 'date' cmd format '%a %b %d %H:%M:%S %Z %Y' works but
        # doesn't get TZ?

        # Could cleanup input a bit? remove leading/trailing [ :,-]?
        # Yes, leading/trailing whitespace to match pandas behavior...
        # Actually, pandas doesn't ignore trailing space, it assumes
        # the *current* month/day if they're missing and there's
        # trailing space, or the month is a single, non zero-padded digit.?!
        val = val.strip()

        for fmt in _TS_FORMATS:
            try:
                ts = datetime.strptime(val, fmt)
                break
            except ValueError:
                pass

        try:
            if isinstance(ts, datetime):
                val = ts
                check = False
            else:
                # what else could the type be here?
                raise TypeError(f"expected datetime, got {ts:r}")

        except NameError:
            raise ValueError("failed to parse time: " + repr(val)) from None

    if check and not isinstance(val, datetime):
        return val

    if val.tzinfo is None:
        # is_dst=None raises NonExistent and Ambiguous TimeErrors
        # when appropriate, same as pandas
        val = zone.localize(val, is_dst=None)

    return (val - epoch).total_seconds()


@register(name="time_range")
@dataclass
class TimeRange:
    """
    A search query representing a time range.

    Parameters
    ----------
    since, until: dates gives as timestamp, datetime, or human-friendly string, optional
    timezone : string
        As in, 'US/Eastern'. If None is given, tzlocal is used.

    Examples
    --------
    Any granularity (year, month, date, hour, minute, second) is accepted.

    >>> TimeRange(since='2014')

    >>> TimeRange(until='2019-07')

    >>> TimeRange(since='2014-07-04', until='2020-07-04')

    >>> TimeRange(since='2014-07-04 05:00')

    """

    timezone: str
    since: Optional[float] = None
    until: Optional[float] = None

    def __init__(self, *, timezone=None, since=None, until=None):
        # Stash the raw values just for use in the repr.
        self._raw_since = since
        self._raw_until = until

        if timezone is None:
            import tzlocal

            lz = tzlocal.get_localzone()
            try:
                timezone = lz.key
            except AttributeError:
                timezone = lz.zone
        self.timezone = timezone
        if since is None:
            self.since = None
        else:
            self.since = _normalize_human_friendly_time(since, tz=self.timezone)
        if until is None:
            self.until = None
        else:
            self.until = _normalize_human_friendly_time(until, tz=self.timezone)
        if since is not None and until is not None:
            if self.since > self.until:
                raise ValueError("since must not be greater than until.")

    def __repr__(self):
        return (
            f"{type(self).__name__!s}("
            f"timezone={self.timezone!r}, since={self._raw_since!r}, until={self._raw_until!r})"
        )

    def encode(self):
        return asdict(self)

    @classmethod
    def decode(cls, *, timezone, since=None, until=None):
        return cls(timezone=timezone, since=since, until=until)



import collections.abc
import copy
import functools
import numbers
import operator

from tiled.adapters.utils import IndexCallable
from tiled.client.container import Container
from tiled.client.utils import handle_error
from tiled.queries import Comparison, Eq, Like
from tiled.utils import safe_json_dump

from .bluesky_run import BlueskyRunV2, BlueskyRunV3
from .queries import RawMongo, ScanIDRange, TimeRange, _PartialUID, _ScanID


class CatalogOfBlueskyRuns(Container):
    """
    This adds some bluesky-specific conveniences to the standard client Container.

    >>> catalog.scan_id[1234]  # scan_id lookup
    >>> catalog.uid["9acjef"]  # (partial) uid lookup
    >>> catalog[1234]  # automatically do scan_id lookup for positive integer
    >>> catalog["9acjef"]  # automatically do (partial) uid lookup for string
    >>> catalog[-5]  # automatically do catalog.values()[-N] for negative integer
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scan_id = IndexCallable(self._lookup_by_scan_id)
        self.uid = IndexCallable(self._lookup_by_partial_uid)
        self._v1 = None

    def __repr__(self):
        # This is a copy/paste of the general-purpose implementation
        # tiled.adapters.utils.tree_repr
        # with some modifications to extract scan_id from the metadata.
        sample = self.items()[:10]
        # Use scan_id (int) if defined; otherwise fall back to uid.
        sample_reprs = [repr(value.metadata.get("start", {}).get("scan_id", key)) for key, value in sample]
        out = "<Catalog {"
        # Always show at least one.
        if sample_reprs:
            out += sample_reprs[0]
        # And then show as many more as we can fit on one line.
        counter = 1
        for sample_repr in sample_reprs[1:]:
            if len(out) + len(sample_repr) > 60:  # character count
                break
            out += ", " + sample_repr
            counter += 1
        approx_len = operator.length_hint(self)  # cheaper to compute than len(node)
        # Are there more in the node that what we displayed above?
        if approx_len > counter:
            out += f", ...}} ~{approx_len} entries>"
        else:
            out += "}>"
        return out

    @property
    def v1(self):
        "Accessor to legacy interface."
        if self._v1 is None:
            from databroker.v1 import Broker

            self._v1 = Broker(self)
            self._v1._version = "1.0"
        return self._v1

    @property
    def v2(self):
        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV2)
        return CatalogOfBlueskyRuns(self.context, item=self.item, structure_clients=structure_clients)

    @property
    def v3(self):
        if not self.is_sql:
            raise NotImplementedError("v3 is only available for SQL-based catalogs.")

        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV3)
        return CatalogOfBlueskyRuns(self.context, item=self.item, structure_clients=structure_clients)

    @functools.cached_property
    def is_sql(self):
        for spec in self.specs:
            if spec.name == "CatalogOfBlueskyRuns":
                if spec.version and spec.version.startswith("3."):
                    return True
                return False

    def __getitem__(self, key):
        # For convenience and backward-compatiblity reasons, we support
        # some "magic" here that is helpful in an interactive setting.
        if isinstance(key, str):
            # CASE 1: Interpret key as a uid or partial uid.
            if len(key) == 36:
                # This looks like a full uid. Try direct lookup first.
                try:
                    return super().__getitem__(key)
                except KeyError:
                    # Fall back to partial uid lookup below.
                    pass
            return self._lookup_by_partial_uid(key)
        elif isinstance(key, numbers.Integral):
            if key > 0:
                # CASE 2: Interpret key as a scan_id.
                return self._lookup_by_scan_id(int(key))
            else:
                # CASE 3: Interpret key as a recently lookup, as in
                # `catalog[-1]` is the latest entry.
                key = int(key)
                return self.values()[key]
        elif isinstance(key, slice):
            if (key.start is None) or (key.start >= 0):
                raise ValueError(
                    "For backward-compatibility reasons, slicing here "
                    "is limited to negative indexes. "
                    "Use .values() to slice how you please."
                )
            return self.values()[key]
        elif isinstance(key, collections.abc.Iterable):
            # We know that isn't a str because we check that above.
            # Recurse.
            return [self[item] for item in key]
        else:
            raise ValueError("Indexing expects a string, an integer, or a collection of strings and/or integers.")

    def _lookup_by_scan_id(self, scan_id):
        results = self.search(Eq("start.scan_id", scan_id))
        if not results:
            raise KeyError(f"No match for scan_id={scan_id}")
        else:
            # Return latest match.
            return results.values().last()

    def _lookup_by_partial_uid(self, partial_uid):
        if len(partial_uid) < 5:
            raise ValueError(f"Partial uid {partial_uid!r} is too short. It must include at least 5 characters.")
        if self.is_sql:
            query = Like("start.uid", f"{partial_uid}%")
        else:
            query = _PartialUID(partial_uids=[partial_uid])
        results = self.search(query).values().head(2)
        if len(results) > 1:
            raise ValueError(
                f"Partial uid {partial_uid} has multiple matches. Include more characters to get a unique match."
            )
        if not results:
            raise KeyError(f"No match for partial_uid {partial_uid}")
        # There is one unique result.
        return results[0]

    def get_serializer(self):
        from tiled.server.app import get_root_tree

        if not hasattr(self.context.http_client, "app"):
            raise NotImplementedError("Only works on local application.")
        tree = self.context.http_client.app.dependency_overrides[get_root_tree]()
        return tree.get_serializer()

    def search(self, query):
        # These query types were formerly handled server side by specially-registered
        # queries. Now that are transformed client side into generic queries that
        # come standard with the Tiled server.

        # Some need to be expressed as a chain of queries.
        if isinstance(query, TimeRange):
            result = self
            if query.since:
                result = Container.search(result, Comparison("ge", "start.time", query.since))
            if query.until:
                result = Container.search(result, Comparison("lt", "start.time", query.until))
        # For backward-compatiblity, accept a dict and interpret it as a Mongo
        # query against the 'start' documents.
        elif isinstance(query, _ScanID):
            if len(query.scan_ids) > 1:
                raise ValueError("Search on multiple ScanIDs in one query is no longer supported.")
            (scan_id,) = query.scan_ids
            query = Eq("start.scan_id", int(scan_id))
            result = super().search(query)
        elif isinstance(query, _PartialUID):
            if len(query.partial_uids) > 1:
                raise ValueError("Search on multiple PartialUIDs in one query is no longer supported.")
            (partial_uid,) = query.partial_uids
            if self.is_sql:
                query = Like("start.uid", f"{partial_uid}%")
            else:
                query = _PartialUID(partial_uids=[partial_uid])
            result = super().search(query)
        elif isinstance(query, ScanIDRange):
            ge = Comparison("ge", "start.scan_id", query.start_id)
            lt = Comparison("lt", "start.scan_id", query.end_id)
            result = super().search(ge).search(lt)
        elif isinstance(query, dict):
            query = RawMongo(start=query)
            result = super().search(query)
        else:
            if hasattr(query, "key"):
                if not query.key.startswith("start.") or query.key.startswith("stop."):
                    # Default to searching RunStart document.
                    query = copy.copy(query)
                    query.key = f"start.{query.key}"
            result = super().search(query)
        return result

    def post_document(self, name, doc):
        link = self.item["links"]["self"].replace("/metadata", "/documents", 1)
        response = self.context.http_client.post(link, content=safe_json_dump({"name": name, "doc": doc}))
        handle_error(response)
